assert(juggy, "Juggy_Definitions.lua must be defined first.")

local CVAR_WordFilter = CV_RegisterVar(
    {
        name = "jug_wordfilter_enabled",
        defaultvalue = "On",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local CVAR_WordFilter_PunishType = CV_RegisterVar(
    {
        name = "jug_wordfilter_punishtype",
        defaultvalue = "ShadowMute",
        flags = CV_NETVAR,
        PossibleValue = 
        {
            Block       = juggy.punishType.Block,
            ShadowMute  = juggy.punishType.ShadowMute,
            Kick        = juggy.punishType.Kick,
            Ban         = juggy.punishType.Ban,
            Quit        = juggy.punishType.Quit
        }
    }
)

local CVAR_WordFilter_AdminImmunity = CV_RegisterVar(
    {
        name = "jug_wordfilter_immunity",
        defaultvalue = "On",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local CVAR_WordFilter_UseDefaultBadWordList = CV_RegisterVar(
    {
        name = "jug_wordfilter_usedefaultbadwordlist",
        defaultvalue = "On",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local CVAR_WordFilter_Unleet = CV_RegisterVar(
    {
        name = "jug_wordfilter_unleet",
        defaultvalue = "On",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local CVAR_WordFilter_WholeWordsOnly = CV_RegisterVar(
    {
        name = "jug_wordfilter_wholewordsonly",
        defaultvalue = "Off",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local CVAR_WordFilter_TauntSoundOnLeave = CV_RegisterVar(
    {
        name = "jug_wordfilter_tauntsoundonleave",
        defaultvalue = "Off",
        flags = CV_NETVAR,
        PossibleValue = CV_YesNo
    }
)

local defaultBadWords = 
{
    -- shoutouts to the dark souls 2 banword list
    "trann", -- tranny, trannies
    "fuck trans", -- fuck trans rights
    "fuck trann", -- fuck tranny/ies
    "transexual", -- transexual, transexuals
    "transsexual", -- typo
    "transvestite", -- transvestite, transvestites
    "transsvestite", -- typo
    -- oh man, so many ways to be racist
    "nig", -- no, I don't genuinely believe you're talking about Nigeria
    "spic",
    -- no, I don't genuinely believe you're "out to get a fag" in the middle of a race
    "fag",
    "rape",
    "raping",
    "rapist",
    "retard",
}

juggy.customWordList = {}

juggy.Chat_WordFilter_AddBadWord = function(player, ...)
    if player == nil then return end

    local extraVariables = select("#", ...)
    for i = 1, extraVariables do
        local badword = select(i, ...)

        if type(badword) == "string" then
            badword = $:lower()

        else continue
        end

        table.insert(juggy.customWordList, badword)

        local logMessage = juggy.JUG_LOGNAME.." - Added following word to the custom banlist: "..badword
        CONS_Printf(player, logMessage)

        if juggy.CVAR_Logging.value >= 1 and player ~= server then
            CONS_Printf(server, logMessage)
        end
    end
end

juggy.Chat_WordFilter_RemoveBadWord = function(player, ...)
    if player == nil then return end

    local extraVariables = select("#", ...)
    for i = 1, extraVariables do
        local badword = select(i, ...)

        if type(badword) == "string" then
            badword = $:lower()

        else continue
        end

        local logMessage = juggy.JUG_LOGNAME.." - Could not find bad word to remove from custom list: "..badword
        for key, value in pairs(juggy.customWordList) do
            if value == badword then
                table.remove(juggy.customWordList, key)
                logMessage = juggy.JUG_LOGNAME.." - Removed following word from the custom banlist: "..badword
                break
            end
        end
        
        CONS_Printf(player, logMessage)

        if juggy.CVAR_Logging.value >= 1 and player ~= server then
            CONS_Printf(server, logMessage)
        end
    end
end

juggy.Chat_WordFilter_ListBadWord = function(player, ...)
    local isDefaultListEnabled = (CVAR_WordFilter_UseDefaultBadWordList.value >= 1)
    local tablesToCheck = { juggy.customWordList }
    if isDefaultListEnabled then
        table.insert(tablesToCheck, defaultBadWords)
    end

    local finalTableString = "\130Currently displaying the following banned words:\n\128"

    for _, table in pairs(tablesToCheck) do

        if table == juggy.customWordList then
            if #(juggy.customWordList) > 0 then
                finalTableString = $ + "\t\130The custom banned word list has the following words:\n\128"
            else
                finalTableString = $ + "\t\130The custom banned word list is currently empty.\n\128"
            end
        elseif table == defaultBadWords then
            finalTableString = $ + "\t\130The default banned word list has the following words:\n\128"
        end

        for _, bannedword in pairs(table) do
            finalTableString = $ + "\t\t" + bannedword + "\n"
        end
    end
    finalTableString = $ + "End of list."

    CONS_Printf(player, finalTableString)
end

COM_AddCommand("jug_wordfilter_removewords", juggy.Chat_WordFilter_RemoveBadWord, 1)
COM_AddCommand("jug_wordfilter_addwords", juggy.Chat_WordFilter_AddBadWord, 1)
COM_AddCommand("jug_wordfilter_listwords", juggy.Chat_WordFilter_ListBadWord, 1)
-- Aliases:
COM_AddCommand("jug_wordfilter_removeword", juggy.Chat_WordFilter_RemoveBadWord, 1)
COM_AddCommand("jug_wordfilter_addword", juggy.Chat_WordFilter_AddBadWord, 1)
COM_AddCommand("jug_wordfilter_listword", juggy.Chat_WordFilter_ListBadWord, 1)

-- By RetroStation in/for ChatBlaster - used with permission
-- Readapted and remade to fit my coding standards better
local function unL33TMessage(message)
    local messageLower = message:lower()

    local charactersToFindAndReplaceWith =
    {
        ["4"] = "a",
        ["@"] = "a",
        ["3"] = "e",
        ["&"] = "e",
        ["1"] = "i",
        ["!"] = "i",
        ["0"] = "o",
        ["8"] = "b",
        ["9"] = "g",
        ["#"] = "h",
        ["5"] = "s",
        ["$"] = "s",
        ["7"] = "t"
    }

    for key, value in pairs(charactersToFindAndReplaceWith) do  
        if messageLower:find(key) then
            -- specific handling for the special character $
            if key == "$" then key = "%"..key end

            messageLower = $:gsub(key, value)
            -- print(messageLower)
        end
    end

    return messageLower
end

juggy.Chat_WordFilter = function(parameters)
    -- Spam filter-related
    if CVAR_WordFilter.value > 0 then

        local isThereMatch = nil

        local messagelower = parameters.message:lower()
        if CVAR_WordFilter_Unleet.value >= 1 then
            messagelower = unL33TMessage(parameters.message)
        end 

        local tablesToCheck = { juggy.customWordList }

        if CVAR_WordFilter_UseDefaultBadWordList.value >= 1 then
            table.insert(tablesToCheck, defaultBadWords)
        end

        local messageWordTable, messageWordAmount = nil, nil;
        if CVAR_WordFilter_WholeWordsOnly.value >= 1 then
            messageWordTable, messageWordAmount = juggy.turnCharDelimitedStringIntoTable(parameters.message, ' ')
        end

        -- literally no message to process, dip
        if CVAR_WordFilter_WholeWordsOnly.value >= 1 and essageWordAmount == 0 then
            return parameters
        end

        for _, table in pairs(tablesToCheck) do
            for _, bannedword in pairs(table) do
                if CVAR_WordFilter_WholeWordsOnly.value >= 1 then
                    for _, wholeword in pairs(messageWordTable) do
                        if wholeword == bannedword or (wholeword.."s") == bannedword then
                            
                            isThereMatch = bannedword
                            break
                        end
                    end
                elseif messagelower:find(bannedword) ~= nil then
                    isThereMatch = bannedword
                    -- print("found it " + messagelower:find(bannedword))
                    break
                end

                -- print(messagelower + " = " + bannedword)
            end

            if isThereMatch ~= nil then break end
        end

        if not (CVAR_WordFilter_AdminImmunity.value >= 1 and
            juggy.IsPlayerServerOrAdmin(parameters.player)) and
            isThereMatch ~= nil then

            if parameters.bannedword == nil then
                parameters.bannedword = isThereMatch
            end

            -- if spam filter is also enabled, this also counts.
            -- this is to prevent spamming multiple messages at once.
            if not (juggy.CVAR_SpamFilter_AdminImmunity.value >= 1 and
                juggy.IsPlayerServerOrAdmin(parameters.player)) and
                juggy.CVAR_SpamFilter.value >= 1 then
                parameters.player.jug_spamtime = juggy.CVAR_SpamFilter_SpamTime.value
            end
            
            if CVAR_WordFilter_PunishType.value == juggy.punishType.Block then
                chatprintf(parameters.player, juggy.filterErrors.filteredWord, false)
                
                parameters.returntype = true
                return parameters

            elseif CVAR_WordFilter_PunishType.value == juggy.punishType.ShadowMute then
                parameters.shadowmutetype = $ | juggy.shadowMuteType.Filtered
            elseif CVAR_WordFilter_PunishType.value == juggy.punishType.Kick then
                if juggy.CVAR_Logging.value >= 1 then
                    CONS_Printf(server,
                        juggy.JUG_LOGNAME.." - Kicked "..parameters.player.name..
                        " (#"..#(parameters.player)..") for saying a bad word: \""..isThereMatch.."\".")
                end

                if CVAR_WordFilter_TauntSoundOnLeave.value >= 1 then
                    S_StartSound(nil, sfx_chshot)
                end

                COM_BufInsertText(server, "kick "..#(parameters.player).." \"JJT: Bad Words Used\"")

                parameters.returntype = true
                return parameters

            elseif CVAR_WordFilter_PunishType.value == juggy.punishType.Ban then
                if juggy.CVAR_Logging.value >= 1 then
                    CONS_Printf(server,
                        juggy.JUG_LOGNAME.." - Banned "..parameters.player.name..
                        " (#"..#(parameters.player)..") for saying a bad word: \""..isThereMatch.."\".")
                end

                if CVAR_WordFilter_TauntSoundOnLeave.value >= 1 then
                    S_StartSound(nil, sfx_chshot)
                end

                COM_BufInsertText(server, "ban "..#(parameters.player).." \"JJT: Bad Words Used\"")

                parameters.returntype = true
                return parameters

            elseif CVAR_WordFilter_PunishType.value == juggy.punishType.Quit then
                if juggy.CVAR_Logging.value >= 1 then
                    CONS_Printf(server,
                        juggy.JUG_LOGNAME.." - Made "..parameters.player.name..
                        " (#"..#(parameters.player)..") quit for saying a bad word: \""..isThereMatch.."\".")
                end

                if CVAR_WordFilter_TauntSoundOnLeave.value >= 1 then
                    S_StartSound(nil, sfx_chshot)
                end

                COM_BufInsertText(parameters.player, "quit") -- lol, lmao even

                parameters.returntype = true
                return parameters

            end

        end
    end

    return parameters
end

juggy.createChatModule(juggy.Chat_WordFilter, juggy.CHATMOD_MOD)

addHook("NetVars", function(network)

    juggy.customWordList = network(juggy.customWordList)

end)